#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>

#include <string>
#include <cstring>
#include <vector>
#include <map>
#include <iostream>

#include "memearray.h"
#include "memehash.h"
#include "memetype.h"
#include "meme.h"

using namespace std;

#include "y.tab.h"

// TODO: support multiple var declares on one line like "define a,b"

// TODO: unit test in tester for string methods

// TODO: write some documentation

// TODO: objects and custom messages

// TODO: bug fails if last line is a comment with no following newline,mb whenever lacking last newline.

// TODO: make the scanner reentrant

// TODO: handle escape sequences in strings, by parsing with yyinput instead of regex
// TODO: date and time support( timestamp, date/time strings, date time format )

// TODO: make nested functions work instead of always being global.
// TODO: lexical scoping of variables when using nested functions.

// TODO: support environment for read
// TODO: support environment for write

// TODO: create a range primitive/object that handles ranges for for loops without storage.
// TODO: destroy ini() or make it make sense

// TODO: (partly done)add lambda support (shouldn't be hard)

// TODO: Handle multiline strings
// TODO: redo MemeArray using vector
// TODO: Cleanup and redivide source files.

// TODO: closures

// TODO: change to use a mark and sweep GC instead of ref counting.


MemePrimitive *getNullPrimitive(void);
MemePrimitive *getIntegerPrimitive(void);
MemePrimitive *getRealPrimitive(void);
MemePrimitive *getBoolPrimitive(void);
MemePrimitive *getStringPrimitive(void);
MemePrimitive *getArrayPrimitive(void);
MemePrimitive *getHashPrimitive(void);
MemePrimitive *getRefPrimitive(void);
MemePrimitive *getAppenderPrimitive(void);
MemePrimitive *getFuncPrimitive(void);

void RegisterConsole(MemeScope& scope);
void RegisterTypeSupport(MemeScope& scope);
void RegisterFile(MemeScope& scope);

MemeScope globals(NULL);
int gLineCount = 0;

void trace(const char *fmt, ...)
{
#if MEME_TRACE
	va_list ap;
	va_start( ap, fmt );
	fprintf(stderr,"trace: ");
	vfprintf(stderr, fmt, ap );
	fprintf(stderr,"\n");
	va_end(ap);
#endif
}

void warning(const char *fmt, ...)
{
	va_list ap;
	va_start( ap, fmt );
	fprintf(stderr,"warning: ");
	vfprintf(stderr, fmt, ap );
	fprintf(stderr,"\n");
	va_end(ap);
}

void error(const char *fmt, ...)
{
	int lineno = gLineCount;

	va_list ap;
	va_start( ap, fmt );
	fprintf(stderr,"error@line#%d: ",lineno);
	vfprintf(stderr, fmt, ap );
	fprintf(stderr,"\n");
	va_end(ap);

	exit(1); // could be better
}

void countLine(void)
{
	// NOTE: globals bad
	gLineCount++;
}

Value ex(MemeScope& locals, nodeType *p) {
	// INFO: the way we have implemented returns and similiar things (like breaks)
	// we must check during each loop for locals.unwind() so that the loop will not
	// continue when the function is returning.

    if (!p)
    	return Value();
    if( locals.unwind() ) // bailing
    	return locals.getReturn();

    switch(p->type) {
    case typeCon:       return p->con.value;
    case typeReal: 		return p->con.rval;
	case typeBool: 		return Value( p->con.value == 1 );
	case typeNull: 		return Value();
	case typeFunc: 		return Value(); // handled during parsing
	case typeCall: 		{
							// build array of values for arguments
							nodeType *argp = p->call.args;
							vector<Value> args;
							while(argp)
							{
								Value v = ex(locals,argp );
#if MEME_TRACE
								cout << "PUSH: " << v.toString() << endl;
#endif
								args.push_back( v );
								argp = argp->next; // TODO: does not work for nested functions.
							}

							// lookup function
							Value *v = locals.lookup(p->call.name);
							if( !v)
								error("Function not found");
							// that will always be a ref
							if( v->type != valueRef )
								error("internal error");
							Value *actual = (Value *)v->ref_.target();
							if( actual->type != valueFunc )
								error("Not a function");
							// get func from ref.
							MemeFunc *func = (MemeFunc *)actual->ref_.target();
							// and call
							return func->call( args );
						}
	case typeArray: 	{
							Value arr = Value(valueArray,MemeArray::newArray());
							// will be backwards
							nodeType *elem = p->array.elements;
							while( elem )
							{
								arr.getPrimitive()->ins( arr, ex(locals, elem) );
								elem = elem->next;
							}
							return arr;
						}
	case typeStr: 		return Value(valueStr,p->str.ref);
	case typeParam:
						error("internal parse error"); break;
    case typeId:        {
							Value *v = locals.lookup(p->id.name);
							if(v)
								return *v;
							else
								error("unknown identifier %s",p->id.name);
						}
    case typeOpr:
        switch(p->opr.oper) {
        case RETURN:    {
							Value rval = ex( locals, p->opr.op[0] );
							locals.func_return( rval );
							return rval;
						}
        case BREAK:		{
							locals.set_break();
							return Value();
						}
        case WHILE:     {
							locals.clr_break();
							while( !locals.unwind() && ex(locals,p->opr.op[0]).toBool() )
								ex(locals,p->opr.op[1]);
							locals.clr_break();
							return Value();
						}
		case UNTIL: 	{
							locals.clr_break();
							do {
								ex(locals,p->opr.op[0]);
							} while( !locals.unwind() && ! ex(locals,p->opr.op[1]).toBool() ) ;
							locals.clr_break();
							return Value();
		}
        case IF:        if ( ex(locals,p->opr.op[0]).toBool() )
                            ex(locals,p->opr.op[1]);
                        else if (p->opr.nops > 2)
                            ex(locals,p->opr.op[2]);
                        return Value();
		case FOR: 		{
							locals.clr_break();
							if( p->opr.op[0]->type != typeId )
								error("variable expected");
							Value container =  ex(locals,p->opr.op[1]); // assumes all containers are ref types
							int max = container.getPrimitive()->len(container).toInt();
							for( int i = 0; !locals.unwind() && i<max; i++ )
							{
								// assign var in scope
								locals.define(p->opr.op[0]->id.name,
									container.getPrimitive()->idx(container, i ) );
                            	ex(locals,p->opr.op[2]); // body
							}
							locals.clr_break();
							return Value();
						}
        case PRINT:     ex(locals,p->opr.op[0]).println(); return Value();
        case ';':       {
							ex(locals,p->opr.op[0]);
							if( !locals.unwind() )
								return ex(locals,p->opr.op[1]);
							else
								break;
						}
        case DEFINE:	{
							if( p->opr.op[0]->type != typeId )
								error("identifier expected");
							else
							{
								if( p->opr.nops == 2 ) {
									MemeRef dref = locals.define(p->opr.op[0]->id.name, ex(locals,p->opr.op[1]).value() );
									return Value( valueRef, dref );
								} else {
									locals.define(p->opr.op[0]->id.name, Value());
									return Value();
								}
							}
						}
        case '=':       {
							if( p->opr.op[0]->type == typeId )
								// added value() because it was aliases variables in a=b instead of copying.
								return locals.assign(p->opr.op[0]->id.name, ex(locals,p->opr.op[1]).value() );
							else
							{
								Value left = ex(locals,p->opr.op[0]);
								Value right = ex(locals,p->opr.op[1]).value();
								// NOTE: this causes it to implode if right is a ref/variable.

								left.getPrimitive()->lassign(left, right );
								return right;
							}
						}
        case UMINUS:    return -ex(locals,p->opr.op[0]);
        case '+':       return ex(locals,p->opr.op[0]) + ex(locals,p->opr.op[1]);
        case '-':       return ex(locals,p->opr.op[0]) - ex(locals,p->opr.op[1]);
        case '*':       return ex(locals,p->opr.op[0]) * ex(locals,p->opr.op[1]);
        case '/':       return ex(locals,p->opr.op[0]) / ex(locals,p->opr.op[1]);
        case '%':       return ex(locals,p->opr.op[0]) % ex(locals,p->opr.op[1]);
        case '<':       return ex(locals,p->opr.op[0]) < ex(locals,p->opr.op[1]);
        case '>':       return ex(locals,p->opr.op[0]) > ex(locals,p->opr.op[1]);
		case '[': 		{
							Value left = ex(locals,p->opr.op[0]);
							return left.getPrimitive()->idx( left, ex(locals,p->opr.op[1]) );
						}
		case APPEND:    {
							Value *vp = new Value( ex(locals,p->opr.op[0]) );
							return Value(valueAppender, MemeRef( vp ));
						}
        case GE:        return ex(locals,p->opr.op[0]) >= ex(locals,p->opr.op[1]);
        case LE:        return ex(locals,p->opr.op[0]) <= ex(locals,p->opr.op[1]);
        case NE:        return ex(locals,p->opr.op[0]) != ex(locals,p->opr.op[1]);
        case EQ:        return ex(locals,p->opr.op[0]) == ex(locals,p->opr.op[1]);
		case AND:       return ex(locals,p->opr.op[0]).toBool() && ex(locals,p->opr.op[1]).toBool();
		case OR:        return ex(locals,p->opr.op[0]).toBool() || ex(locals,p->opr.op[1]).toBool();
		case NOT:       return ! ex(locals,p->opr.op[0]).toBool();

        }
    }
    return Value();
}

static void storeArgs(int argc,const char **argv)
{
	// Make an Value array from argv and stuff it into globals
	Value arr = Value(valueArray,MemeArray::newArray());
	for(int i=1;i<argc;i++)
		arr.getPrimitive()->cat( arr, Value(valueStr,MemeString::newString( argv[i] ) ) );
	globals.define("ARGV",arr);
}

static void storeEnv(char **envp)
{
	MemeRef menv = MemeHash::newHash();
	MemeHash* mhash = (MemeHash*)menv.target();

	for(char **kv=envp;*kv;kv++) {
		trace("env: %s\n",*kv);
		// Store env values as hash of String -> String
		char *s = *kv;
		char *eq = strstr(s,"=");
		if(eq) {
			mhash->insert(
				Value(valueStr, MemeString::newString(s,eq-s)),
				Value(valueStr, MemeString::newString(eq+1))
			);
		}
	}
	globals.define("ENV",Value( valueHash, menv));
}

void setup(int argc, const char **argv, char **envp)
{
	trace("setup: argc=%d",argc);

	gLineCount = 0;

	Value::registerPrimitive( valueNull, getNullPrimitive() );
	Value::registerPrimitive( valueInt, getIntegerPrimitive() );
	Value::registerPrimitive( valueReal, getRealPrimitive() );
	Value::registerPrimitive( valueBool, getBoolPrimitive() );
	Value::registerPrimitive( valueStr, getStringPrimitive() );
	Value::registerPrimitive( valueArray, getArrayPrimitive() );
	Value::registerPrimitive( valueHash, getHashPrimitive() );
	Value::registerPrimitive( valueRef, getRefPrimitive() );
	Value::registerPrimitive( valueAppender, getAppenderPrimitive() );
	Value::registerPrimitive( valueFunc, getFuncPrimitive() );

	storeArgs(argc,argv);
	storeEnv(envp);

	RegisterConsole(globals);
	RegisterTypeSupport(globals);
	RegisterFile(globals);
}

//////////
#define SIZEOF_NODETYPE ((char *)&p->gaurd - (char *)p)

nodeType *con(int value) {
    nodeType *p;
    size_t nodeSize;

    // allocate node 
    nodeSize = SIZEOF_NODETYPE + sizeof(conNodeType);
    if ((p = (nodeType *)malloc(nodeSize)) == NULL)
        error("out of memory");
	p->next=NULL;

    // copy information 
    p->type = typeCon;
    p->con.value = value;

    return p;
}

nodeType *realval(double value) {
    nodeType *p;
    size_t nodeSize;

    // allocate node
    nodeSize = SIZEOF_NODETYPE + sizeof(conNodeType);
    if ((p = (nodeType *)malloc(nodeSize)) == NULL)
        error("out of memory");
	p->next=NULL;

    // copy information
    p->type = typeReal;
    p->con.rval = value;

    return p;
}

nodeType *nullval(void) {
    nodeType *p;
    size_t nodeSize;

    // allocate node
    nodeSize = SIZEOF_NODETYPE + sizeof(conNodeType);
    if ((p = (nodeType *)malloc(nodeSize)) == NULL)
        error("out of memory");
	p->next=NULL;

	// null value
    p->type = typeNull;

    return p;
}

nodeType *chainArg(nodeType *left, nodeType *right) {
	right->next = left;
	return right;
}

nodeType *newArray(nodeType *args) {
    nodeType *p;
    size_t nodeSize;

    // allocate node
    nodeSize = SIZEOF_NODETYPE + sizeof(conNodeType);
    if ((p = (nodeType *)malloc(nodeSize)) == NULL)
        error("out of memory");
	p->next = NULL;

	// signifies new array
    p->type = typeArray;
	p->array.elements = args;


    return p;
}

nodeType *boolval(bool bval) {
	nodeType *p;

    size_t nodeSize;

    // allocate node
    nodeSize = SIZEOF_NODETYPE + sizeof(conNodeType);
    if ((p = (nodeType *)malloc(nodeSize)) == NULL)
        error("out of memory");
	p->next=NULL;

    // copy information
    p->type = typeBool;
    p->con.bval = bval;

    return p;
}

nodeType *id(const char *name) {
    nodeType *p;
    size_t nodeSize;

    // allocate node
    nodeSize = SIZEOF_NODETYPE + sizeof(idNodeType);
    if ((p = (nodeType *)malloc(nodeSize)) == NULL)
        error("out of memory");
	p->next=NULL;

    // copy information
    p->type = typeId;
    p->id.name = name;

    return p;
}

nodeType *param(const char *name) {
    nodeType *p;
    size_t nodeSize;

    // allocate node
    nodeSize = SIZEOF_NODETYPE + sizeof(idNodeType);
    if ((p = (nodeType *)malloc(nodeSize)) == NULL)
        error("out of memory");
	p->next=NULL;

    // copy information
    p->type = typeParam;
    p->param.name = name;

    return p;
}

nodeType *str(char *text) {
    nodeType *p;
    size_t nodeSize;

    // allocate node
    nodeSize = SIZEOF_NODETYPE + sizeof(strNodeType);
    if ((p = (nodeType *)malloc(nodeSize)) == NULL)
        error("out of memory");
	p->next=NULL;

    // copy information
    p->type = typeStr;
	// p->str.text = text;
	p->str.ref = MemeString::newString( text );


    return p;
}

nodeType *opr(int oper, int nops, ...) {
    va_list ap;
    nodeType *p;
    size_t nodeSize;
    int i;

    // allocate node
    nodeSize = SIZEOF_NODETYPE + sizeof(oprNodeType) +
        (nops - 1) * sizeof(nodeType*);
    if ((p = (nodeType *)malloc(nodeSize)) == NULL)
        error("out of memory");
	p->next=NULL;

    // copy information
    p->type = typeOpr;
    p->opr.oper = oper;
    p->opr.nops = nops;
    va_start(ap, nops);
    for (i = 0; i < nops; i++)
        p->opr.op[i] = va_arg(ap, nodeType*);
    va_end(ap);
    return p;
}
nodeType *makeFunctionNode(const char *name, nodeType *params)
{
	// save the entire tree under impl to the message handler mesgName for class typeName,
	// and return impl
    nodeType *p;
    size_t nodeSize;

    // allocate node
    nodeSize = SIZEOF_NODETYPE + sizeof(conNodeType);
    if ((p = (nodeType *)malloc(nodeSize)) == NULL)
        error("out of memory");
	p->next=NULL;

	// signifies new array
    p->type = typeFunc;
	p->func.name = name;
	p->func.params = params;
	p->func.impl = NULL; // up to caller.
	return p;
}

nodeType *define_function(const char *name,  nodeType *params, nodeType *impl)
{
	nodeType *p = makeFunctionNode( name, params );
	p->func.impl = impl;

	// assign function into global symbols
	MemeFunc *func= new MemeNativeFunc( &globals, p );
	Value fref( valueFunc, MemeRef( func ) );
	globals.define( p->func.name, fref );

	return p;
}

nodeType *callFunction(const char *name, nodeType *args) {
    nodeType *p;
    size_t nodeSize;

    // allocate node
    nodeSize = SIZEOF_NODETYPE + sizeof(conNodeType);
    if ((p = (nodeType *)malloc(nodeSize)) == NULL)
        error("out of memory");
	// p->next=args; // TODO: I think chaining it like this is why nested won't work
	p->next=NULL;
	p->call.args=args;

	p->call.name = name;

	// signifies new array
    p->type = typeCall;

    return p;
}

void freeNode(nodeType *p) {
    int i;

    if (!p) return;
    if (p->type == typeOpr) {
        for (i = 0; i < p->opr.nops; i++)
            freeNode(p->opr.op[i]);
	} else if (p->type == typeId ) {
		if( p->id.name )
			free((void *)p->id.name);
	}
	if( p->next )
		freeNode( p->next );

    free (p);
}
