#include <stdio.h>

#include "memebase.h"

// just placeholder for dispatch
MemeBase::MemeBase(void)
{
}
MemeBase::~MemeBase()
{
}

// int refCount_

MemeCollect::MemeCollect(void)
	: refCount_(0), untracked_(false)
{
}



MemeCollect::~MemeCollect()
{
}

int MemeCollect::refCount(void)
{
	return refCount_;
}

int MemeCollect::incRefCount(void)
{
	return ++refCount_;
}


void MemeCollect::decRefCount(void)
{

	refCount_--;
}

void MemeRef::newRef(MemeCollect *obj )
{
	// if old ref, dec it
	if( target_ )
		target_->decRefCount();

	target_ = obj;


	// inc new ref,  and if brand new add it to the gc pool
	if( target_ && target_->incRefCount() == 1 && ! target_->untracked_ )
	{
		MemeReaper *reaper = MemeReaper::Reaper();
		if( reaper )
			reaper->track( target_);
	}
}


MemeRef::MemeRef(MemeCollect* obj)
	: target_(NULL)
{
	newRef( obj );
}


MemeRef::MemeRef(const MemeRef& mr)
	: target_(NULL)
{
	newRef( mr.target_ );
}


MemeRef::~MemeRef()
{
	if( target_ )
		target_->decRefCount();
}


const MemeRef& MemeRef::operator = (const MemeRef& mr)
{
	newRef( mr.target_ );
	return *this;
}

void MemeRef::setTarget(MemeCollect *obj)
{
	newRef( obj );
}

MemeReaper *MemeReaper::sReaper = NULL;

// std::list<MemeCollect*> mHeap
 // as in garbage, not as in a heap heap
MemeReaper::MemeReaper(void)
{
	sReaper = this;
}


MemeReaper::~MemeReaper()
{
	reap();
	if( sReaper == this )
		sReaper = NULL;
}



void MemeReaper::track( MemeCollect *mc )
{
	static int modit = 0;
	pile_.push_front( mc );
	// every so often, clean some stuff up
	if( (modit & 0xff) == 0 )
		reap();
	modit++;
}


void MemeReaper::reap(void)
{
	std::list<MemeCollect*>::iterator it = pile_.begin();
	while( it != pile_.end() )
	{
		if( (*it)->refCount() == 0 )
		{
			delete (*it);
			it = pile_.erase( it );
		}
		else
			it++;
	}
}

class TestCollect : public MemeCollect {
	private:
		bool& flag_;
	public:
		TestCollect(bool& f)
			: flag_(f)
		{
			flag_ = false;
		}
		~TestCollect()
		{
			flag_ = true;
		}
};


// gc unit test
bool MemeReaper::test(void)
{
	bool flagx, flagy;
	MemeReaper reaper;
	TestCollect *x = new TestCollect(flagx);
	TestCollect *y = new TestCollect(flagy);
	{
		reaper.reap(); // should not reap not in pool yet
		if( flagx || flagy ) return false;
		MemeRef x1(x);

		if( flagx || flagy ) return false; // x = 1, y not in pool
		{
			MemeRef x2 = x1;
			MemeRef y1(y);
			reaper.reap(); // all nonzero
			if( flagx || flagy ) return false; // x = 2, y = 1
		}
		// reap y
		reaper.reap();
		if( flagx || ! flagy ) return false;
	}
	// reap x
	reaper.reap();
	if( ! flagx ) return false;

	return true;
}

