/*
 * memehash.cpp
 *
 *  Created on: Mar 8, 2009
 *      Author: richardanton
 */

#include "meme.h"
#include "memehash.h"
#include <iostream>
#include <cstring>
#include <cstdlib>
#include <cstdio>

MemeHash::MemeHash() {

}

MemeHash::~MemeHash() {

}

std::string MemeHash::hkey(const Value& val)
{
	return string( val.toString() );
}

Value& MemeHash::operator[] (const Value& key)
{
	return idx( key );
}

Value& MemeHash::idx(const Value& key)
{
	std::map<std::string,Value>::iterator it = map_.find( hkey(key) );
	if( it == map_.end() )
	{
		// NOTE: we need to make sure this can never
		// have side effects, it's somewhat experimental.
		//map_.insert( MemeHashEntry( hkey(key), Value() ) );
		insert( key, Value() );
		it = map_.find( hkey(key) );

		/// NOTE: what are we going to return here.
		// error("key not found");
	}
	// make sure nobody gets confused and tries to delete it
	it->second.untracked();

	return it->second;
}

size_t MemeHash::count( void )
{
	return map_.size();
}

void MemeHash::insert( const Value& key, const Value& val )
{
	// NOTE: should really be called set
	std::map<std::string,Value>::iterator it = map_.find( hkey(key) );
	if( it == map_.end() )
		map_.insert( MemeHashEntry( hkey(key), val ) );
	else
		it->second = val;
}

Value MemeHash::erase( const Value& key )
{
	Value result;
	std::map<std::string,Value>::iterator it = map_.find( hkey(key) );
	if( it != map_.end() )
	{
		result = it->second;
		map_.erase( it );
	}
	return result;
}

bool MemeHash::exists( const Value& key )
{
	std::map<std::string,Value>::iterator it = map_.find( hkey(key) );
	return ( it != map_.end() );
}

// convert to string
Value MemeHash::toString(void)
{
	MemeString *ms = new MemeString( "[" );
	std::map<std::string,Value>::iterator it;
	for( it= map_.begin(); it != map_.end(); it++ )
	{
		if( it != map_.begin() )
			ms->append(",");
		ms->append( it->first.c_str() );
		ms->append(":");
		ms->append( it->second.toString() );
	}
	ms->append( "]" );
	return Value(valueStr, MemeRef(ms));
}

// creates a new ref counted hash
MemeRef MemeHash::newHash(void)
{
	MemeHash *hash = new MemeHash();
	if( !hash ) error( "out of memory" );
	return MemeRef( hash );
}

// set array to hold all the keys
Value MemeHash::keys(void)
{
	MemeArray *arr = new MemeArray();
	std::map<std::string,Value>::iterator it;
	for( it= map_.begin(); it != map_.end(); it++ )
	{
		arr->push( Value( valueStr, MemeString::newString( it->first.c_str() ) ) );
	}
	return Value(valueArray, MemeRef(arr));
}

// set array to hold all the values
Value MemeHash::values(void)
{
	MemeArray *arr = new MemeArray();
	std::map<std::string,Value>::iterator it;
	for( it= map_.begin(); it != map_.end(); it++ )
	{
		arr->push( Value( valueStr, MemeString::newString( it->second.toString() ) ) );
	}
	return Value(valueArray, MemeRef(arr));
}


// TODO: NOTE: we are going to need to be able to return a special primitive
// from undefined idx and [] operations that allows assigning to a missing
// index via lassign() but always returns error or null if dereferenced.

bool MemeHash::test(void)
{
	// new, insert, idx, [], erase, count, exists
	// idx by str and int
	// assign via idx and []
	// keys, values
	MemeHash hash;

	Value fooKey = Value(valueStr, MemeString::newString("foo"));
	Value barKey = Value(valueStr, MemeString::newString("bar"));
	Value numKey = Value(111);
	Value newKey = Value(valueStr, MemeString::newString("new"));


	hash.insert( fooKey, Value(42) );
	hash.insert( barKey, Value(69) );
	hash.insert( numKey, Value(111) );

	if( hash.count() != 3 )
		return false;

	Value v;
	v = hash[ fooKey ];
	if( v.toInt() != 42 )
		return false;
	v = hash[ barKey ];
	if( v.toInt() != 69 )
		return false;
	v = hash[ numKey ];
	if( v.toInt() != 111 )
		return false;

	Value ass = hash.toString();
	// std::cout << MemeString::strFromRef( ass.ref_ ) << "\n";

	// test doesn't work so well since order is not guaranteed
	//if( ass != Value(valueStr,MemeRef(MemeString::newString("Hash[111:111,bar:69,foo:42]"))) )
	//	return false;

	// insert via []
	hash[newKey] =  Value(valueStr,MemeRef(MemeString::newString("shiny")));
	if( hash.count() != 4 )
		return false;
	const char *cstr = hash[newKey].toString();
	if( strcmp( cstr, "shiny" ) )
		return false;

	hash.erase( barKey );

	if( hash.count() != 3 )
		return false;
	Value fooVal = hash.idx(fooKey);
	// std::cout << "foo=>" << fooVal.toString() << "\n";
	if( hash[fooKey].toInt() != 42 )
		return false;

	Value numVal = hash.idx(numKey);
	// std::cout << "111=>" << numVal.toString() << "\n";
	if( hash[numKey].toInt()  != 111 )
		return false;

	if( !hash.exists( fooKey ) )
		return false;
	// std::cout << "foo exists\n";

	if( hash.exists( barKey ) )
		return false;
	// std::cout << "bar does not exist\n";

	return true;
}

