

#include "memetype.h"
#include "meme.h"

class RefPrimitive : public MemePrimitive {
	private:
		// deref for value ptr
		Value *deref(const Value& v);
	public:
		// these are not abstract, they all throw
		// errors if called on an object that does
		// not implement
		RefPrimitive();
		~RefPrimitive();

		virtual Value ini(const Value& x);

		virtual Value add(const Value& a, const Value& b);
		virtual Value sub(const Value& a, const Value& b);
		virtual Value neg(const Value& x);

		virtual Value mul(const Value& a, const Value& b);
		virtual Value div(const Value& a, const Value& b);
		virtual Value mod(const Value& a, const Value& b);

		virtual Value cmp(const Value& a, const Value& b);

		virtual Value str(const Value& a);
		virtual Value num(const Value& a);
		
		virtual Value len(const Value& a);
		virtual Value idx(const Value& a, const Value& i);
		virtual Value cat(const Value& a, const Value& i);

		virtual void lassign(const Value& left, const Value& right);
};

MemePrimitive *getRefPrimitive(void)
{
	return new RefPrimitive();
}

Value *RefPrimitive::deref(const Value& v) 
{
	Value *ptr = (Value *)v.ref_.target();
	return ptr;
}

// these are not abstract, they all throw
// errors if called on an object that does
// not implement
RefPrimitive::RefPrimitive()
{
}

RefPrimitive::~RefPrimitive()
{
}

Value RefPrimitive::ini(const Value& x)
{
	return x;
}

void RefPrimitive::lassign(const Value& left, const Value& right)
{
	deref(left)->assign( right );
}

Value RefPrimitive::add(const Value& a, const Value& b)
{
	Value *d=deref(a);
	return d->getPrimitive()->add( *d, b );
}

Value RefPrimitive::sub(const Value& a, const Value& b)
{
	Value *d=deref(a);
	return d->getPrimitive()->sub( *d, b );
}

Value RefPrimitive::neg(const Value& x)
{
	Value *d=deref(x);
	return d->getPrimitive()->neg( *d );
}

Value RefPrimitive::mul(const Value& a, const Value& b)
{
	Value *d=deref(a);
	return d->getPrimitive()->mul( *d, b );
}

Value RefPrimitive::div(const Value& a, const Value& b)
{
	Value *d=deref(a);
	return d->getPrimitive()->div( *d, b );
}

Value RefPrimitive::mod(const Value& a, const Value& b)
{
	Value *d=deref(a);
	return d->getPrimitive()->mod( *d, b );
}

Value RefPrimitive::cmp(const Value& a, const Value& b)
{
	Value *d=deref(a);
	return d->getPrimitive()->cmp( *d, b );
}

Value RefPrimitive::str(const Value& a)
{
	Value *d=deref(a);
	return d->getPrimitive()->str( *d );
}

Value RefPrimitive::num(const Value& a)
{
	Value *d=deref(a);
	return d->getPrimitive()->num( *d );
}

Value RefPrimitive::len(const Value& a)
{
	Value *d=deref(a);
	return d->getPrimitive()->len( *d );
}

Value RefPrimitive::idx(const Value& a, const Value& i)
{
	Value *d=deref(a);
	return d->getPrimitive()->idx( *d, i );
}

Value RefPrimitive::cat(const Value& a, const Value& i)
{
	Value *d=deref(a);
	return d->getPrimitive()->cat( *d, i );
}
